(* $Id: fork.ml,v 1.3 2010-10-01 12:58:28 qcarbonn Exp $ *)

(* Fork is a simple program, feed it with a valid address
  (fff.ooo.rrr.kkk:port), it will connect to this address and then
  split what it receives in two streams, sending them to the two
  addresses it gives back through its stdout. *)

open Unix;;

let spawn f =
  if fork () = 0 then f ()
  else ()
;;
(** Launch the function [f] in parallel. *)

let dynbind fd =
  bind fd (ADDR_INET (inet_addr_any, 0));
  getsockname fd
;;
(** Dynamically find a free address (pair (ip, port)) and bind it to
  [fd]. *)

let server ic sock1 sock2 =
  let rec send oc1 oc2 =
    match Marshal.from_channel ic with
    | None ->
      Marshal.to_channel oc1 None [];
      Marshal.to_channel oc2 None []
    | Some x ->
      Marshal.to_channel oc1 (Some x) [];
      send oc2 oc1 in
  listen sock1 3; listen sock2 3;
  let (sock1, _addr1), (sock2, _addr2) =
    (accept sock1),
    (accept sock2) in
  let oc1, oc2 =
    (out_channel_of_descr sock1),
    (out_channel_of_descr sock2) in
  send oc1 oc2;
  exit 0
;;
(** Run a fork server, sending alternatively data to [sock1] and
  [sock2] once they are connected. *)

let fork ic =
  let fd1, fd2 =
    (socket PF_INET SOCK_STREAM 0),
    (socket PF_INET SOCK_STREAM 0) in
  let addr1, addr2 =
    (dynbind fd1),
    (dynbind fd2) in
  spawn (fun () -> server ic fd1 fd2);
  (addr1, addr2)
;;
(** Start a server forking data given in the input channel [ic] into
  two streams sent to the tuple of addresses it returns. *)

let print_sockaddr oc addr =
  match addr with
  | ADDR_UNIX addr -> Printf.fprintf oc "%s" addr
  | ADDR_INET (addr, port) ->
    Printf.fprintf oc "%s:%d" (string_of_inet_addr addr) port
;;
(** Display a sockaddr [addr] in the output channel [oc]. *)

let read_sockaddr ic =
  let sb = Scanf.Scanning.from_channel ic in
  let addr, port =
    Scanf.bscanf sb "%s@:%d\n"
      (fun a p ->
        (inet_addr_of_string a), p) in
  ADDR_INET (addr, port)
;;
(** Retreive an inet address read from the channel [ic]. *)

let open_input_channel config_chan =
  let sock = socket PF_INET SOCK_STREAM 0 in
  let target = read_sockaddr config_chan in
  connect sock target;
  in_channel_of_descr sock
;;
(** Connect to the address read from the input channel [config_chan]
  and return an [in_channel] ready to use. *)

let main () =
  let ic = open_input_channel Pervasives.stdin in
  let addr1, addr2 = fork ic in
  Printf.printf "(%a, %a)\n" print_sockaddr addr1 print_sockaddr addr2;
  Pervasives.flush Pervasives.stdout
;;

main ();;
