(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklrun.ml,v 1.15.2.1 2011-09-21 13:08:04 fclement Exp $ *)

type ('a, 'b) either = Left of 'a | Right of 'b;;

module type PARSERS = sig
  type conf_entry = Int of int | Str of string;;
  val parse_conf :
    string -> ((string * conf_entry) list list * string) option;;
  val parse_spawnd_port : string -> int option;;
  val explode_shell_command :
    string -> string -> string -> string list option;;
end
;;

module Parsers : PARSERS = struct
   open Parser;;
   open Parser_plus;;

  let non_escape = any (List.map char (notc ['\\'; '"';]));;
                                  (* helping emacs '"' *)
  let in_comm = any (List.map char (notc ['('; '*';]));;
  let p_space = ignore p_space;;

  let p_comment = (* parsing of nested comments *)
    let rec p_comment () =
      let rec p_comment_in () =
        many in_comm >>= fun _ ->
          (p_comment () >> p_comment_in ()) +++
          (p_word "*)" +++ eof >> return ()) +++
          ((char '(' +++ char '*') >> p_comment_in ()) in
      p_word "(*" >> p_comment_in () in
    p_comment ()
  ;;

  let p_spaces = many (p_comment +++ p_space);;

  let code_to_char l =
    Char.chr (to_num l)
  ;;

  let p_string =
    let p_escape k =
      char '\\' >>
          (times 3 p_digit >>= (fun l -> k (code_to_char l))) +++
          (char '"' >>= fun _ -> k '"') in
    let rec p_str_body k =
      many non_escape >>= (fun x ->
          (p_escape (fun c -> p_str_body (fun y -> k (x @ (c :: y))))) +++
          (char '"' >> k x)) in
    char '"' >> p_str_body return
  ;;

  let rec p_spaced p =
    p_spaces >> (p >>= (fun x -> p_spaces >> return x))
  ;;

  let p_entry p =
    p_spaced (p_ident) >>= fun ident ->
    p_spaced (char '=') >> (
    p >>= fun x ->
    p_spaced (maybe (char ';')) >> (
    return (ident, x)))
  ;;

  type conf_entry = Int of int | Str of string;;

  let p_entry =
    let p_value =
      (p_integer >>= (fun x -> return (Int x))) +++
      (p_string >>= (fun x -> return (Str (implode x)))) in
    p_entry p_value
  ;;

  let p_host =
    char '{' >> (
    many p_entry >>= fun l ->
    char '}' >> ( return l ))
  ;;

  let parse_conf = full_parse (many (p_spaced p_host));;

  let parse_spawnd_port =
    let p_port =
      let p_not_colon = any (List.map char (notc [':'])) in
      let till_colon = many p_not_colon >> char ':' in
      (till_colon >> till_colon) >> p_integer in
    parse p_port
  ;;

  let explode_shell_command host command =
    let p_in_word = any (List.map char (notc ('\'' :: '%' :: spaces))) in
    let p_in_quote = any (List.map char (notc ['%'; '\''])) in
    let p_escape =
      (char 'h' >> return host) +++
      (char 'c' >> return command) +++
      (char '%' >> return "%") +++
      (char '\'' >> return "'") in
    let rec p_quote () =
      many p_in_quote >>= fun b ->
      let b = implode b in
      (char '%' >> (p_escape >>= fun esc ->
      p_quote () >>= fun e ->
      return (b ^ esc ^ e))) +++
      (char '\'' >> return b) in
    let p_end_of_word ret =
      ignore (many1 p_space) +++ eof >> return ret in
    let rec p_percent_word () =
      many p_in_word >>= fun b ->
      let b = implode b in
      (char '%' >> (p_escape >>= fun esc ->
      p_percent_word () >>= fun e ->
      return (b ^ esc ^ e))) +++
      (char '\'' >> (p_quote () >>= fun q ->
      p_percent_word () >>= fun e ->
      return (b ^ q ^ e))) +++
      p_end_of_word b
    in
    let p_cmd = many_till eof (p_percent_word ()) in
    parse p_cmd
  ;;

end
;;

type host_entry = {
  color : int;
  host : string;
  name : string;
  shell : string;
  skl_path : string option;
  spawnd : string;
};;

let default_node =
  { shell = "sh -c '%c'"
  ; spawnd = "spawnd"
  ; skl_path = None
  ; name = "default node"
  ; color = 2
  ; host = "127.0.0.1"
  }
;;

let mk_node warn settings =
  let h = Hashtbl.create 10 in
  let findo k =
    try Some (Hashtbl.find h k)
    with Not_found -> None in
  let get_int k d =
    match findo k with
    | Some (Parsers.Int i) -> i
    | Some (Parsers.Str _) ->
      warn (k ^ " was expected of type int"); d
    | None -> d in
  let get_str k d =
    match findo k with
    | Some (Parsers.Int _) ->
      warn (k ^ " was expected of type string"); d
    | Some (Parsers.Str s) -> s
    | None -> d in
  let get_str_opt k =
    match findo k with
    | Some (Parsers.Str s) -> Some s
    | None -> None
    | Some (Parsers.Int _) ->
      warn (k ^ " was expected of type string"); None in
  let add_warn (k, v) =
    if Hashtbl.mem h k then warn (k ^ " is duplicated") else
    Hashtbl.add h k v in
  List.iter add_warn settings;
  let host = default_node in
  let host = { host with color = get_int "color" host.color; } in
  let host = { host with host = get_str "host" host.host; } in
  let host = { host with name = get_str "name" host.name; } in
  let host = { host with shell = get_str "shell" host.shell; } in
  let host = { host with skl_path = get_str_opt "skl_path"; } in
  let host = { host with spawnd = get_str "spawnd" host.spawnd; } in
  host
;;

(* FIXME unused *)
(* let mk_shell_command shell host command = *)
(*   let string_from s i = *)
(*     String.sub s i (String.length s - i) in *)
(*   let rec loop shell = *)
(*     if String.contains shell '%' then begin *)
(*       let index = String.index shell '%' in *)
(*         match shell.[index + 1] with *)
(*         | 'h' | 'c' as c -> *)
(*           String.sub shell 0 index ^ *)
(*           (if c = 'h' then host else command) ^ *)
(*           loop (string_from shell (index + 2)) *)
(*         | _ -> *)
(*           String.sub shell 0 (index + 1) ^ *)
(*           loop (string_from shell (index + 1)) *)
(*     end else shell in *)
(*   loop shell *)
(* ;; *)

let explode_command cmd =
  let string_from s i =
    String.sub s i (String.length s - i) in
  let is_space c = List.mem c Parser_plus.spaces in
  let rec skip_spaces s =
    if s = "" then s else
    if is_space s.[0]
    then skip_spaces (string_from s 1)
    else s in
  let rec loop s =
    let s = skip_spaces s in
    if List.exists (String.contains s) Parser_plus.spaces then
      let index =
        List.fold_left
          (fun i c ->
            min i (if String.contains s c then String.index s c else i))
          (String.length s)
          Parser_plus.spaces in
      String.sub s 0 index :: loop (string_from s index)
    else if s = "" then [] else [s] in
  Array.of_list (loop cmd)
;;

exception Expand_error of string;;
let expand_error s = raise (Expand_error s);;

let compile_slave_nodes master_addr default_skl_path nodes =
  let in_compile b node =
    let fail_action =
      Batch.Fail_error
        (Printf.sprintf "Failed to launch slave %S" node.name) in
    let spawnd_command =
      let skl_path =
        match node.skl_path with
        | Some p -> p
        | None -> default_skl_path in
      Printf.sprintf "%s -sklpath %s -color %d -slave %s"
        node.spawnd skl_path node.color master_addr in
    let cmd =
      match
        Parsers.explode_shell_command
          node.host spawnd_command node.shell
      with
      | Some cmd -> Array.of_list cmd
      | None -> expand_error node.shell in
    if Array.length cmd > 0 then
      Batch.add_shell_task b fail_action
        { Batch.program = cmd.(0); Batch.args = cmd; } Batch.Background
    else b in
  let b = Batch.empty in List.fold_left in_compile b nodes
;;
(* Builds the batches that will run the commands to launch the remote
   slave spawnd daemons. *)

let compile_master_node nnodes node =
  let b = Batch.empty in
  let fail_action =
    Batch.Fail_error
      (Printf.sprintf "Failed to start master node (%S)" node.name) in
  let spawnd_command =
    Printf.sprintf "%s -master %d" node.spawnd nnodes in
  let cmd =
    match
      Parsers.explode_shell_command
        node.host spawnd_command node.shell
    with
    | Some cmd -> Array.of_list cmd
    | None -> expand_error node.shell in
  if Array.length cmd > 0 then
    Batch.add_shell_task b fail_action
      { Batch.program = cmd.(0); Batch.args = cmd; } Batch.Background
  else b
;;
(* Buids the batch that will launch the master spawnd program. *)

let rec skip_n_input_line ic n t =
  if n = 0 then Tio.input_line ic t else
  match Tio.input_line ic t with
  | Tio.Line _ -> skip_n_input_line ic (n - 1) t
  | Tio.Eof | Tio.Timeout as err -> err
;;

exception Parse_error of string;;
let parse_error m = raise (Parse_error m);;

let mk_batch warn conf local_skl_path skl_code =
  let sample s =
    let take_n s n =
      let n = min (String.length s) n in
      String.sub s 0 n in
    take_n s (Config.get_err_length ()) ^ "..." in
  match Parsers.parse_conf conf with
  | Some (nodes, "") ->
    let (master_node, slave_nodes) =
      match List.map (mk_node warn) nodes with
      | master_node :: slave_nodes ->
        let slave_nodes =
          if slave_nodes = [] then [default_node] else slave_nodes in
        (master_node, slave_nodes)
      | [] -> (default_node, [ default_node ]) in
    let nslaves = List.length slave_nodes in
    let master_batch =
      compile_master_node nslaves master_node
    and slave_batches ic =
      let tmout = Config.get_timeout () in
      let ic = Tio.tic_of_in_channel ic in
      let line =
        match Tio.input_line ic tmout with
        | Tio.Line l -> l
        | Tio.Eof | Tio.Timeout -> "" in
      match Parsers.parse_spawnd_port line with
      | Some port ->
        let master_addr = Printf.sprintf "%s:%d" master_node.host port in
        let b_slaves =
          compile_slave_nodes master_addr local_skl_path slave_nodes in
        let run_skl_code pids =
          let b = Batch.empty in
          let dead_proc pid = fst (Unix.waitpid [Unix.WNOHANG] pid) > 0 in
          if List.exists dead_proc pids then
            Batch.add_fail_task b
              "One slave died before the start of the computation, aborting"
          else begin
            match skip_n_input_line ic nslaves tmout with
            | Tio.Line line ->
              begin match Parsers.parse_spawnd_port line with
              | Some port ->
                let fail_action = Batch.Fail_error "Client code failed" in
                Batch.add_shell_task b fail_action
                  (skl_code master_node.host port) Batch.Normal
              | None ->
                let err_str =
                  Printf.sprintf "Spawnd sent invalid data : %S, aborting" line in
                Batch.add_fail_task b err_str
              end
            | Tio.Eof ->
              let err_str =
                "Spawnd exited prematurely, aborting" in
              Batch.add_fail_task b err_str
            | Tio.Timeout ->
              let err_str =
                "Waiting for spawnd response for too long, aboting" in
              Batch.add_fail_task b err_str
          end in
        let fail_action =
          Batch.Fail_error "Failed to launch the client sklml code" in
        Batch.add_control_task b_slaves fail_action run_skl_code
      | None ->
        let b = Batch.empty in
        let err_str = Printf.sprintf "Spawnd sent invalid data : %S" line in
        Batch.add_fail_task b err_str in
    let b = Batch.empty in
    let fail_action =
      Batch.Fail_error "Parallel network failed" in
    Batch.add_pipe_task b fail_action master_batch slave_batches
  | Some (_, s) -> parse_error (sample s)
  | None -> parse_error (sample conf)
;;

let parse_argv argv =
  let (append_cmd, get_cmd) =
    let cmdl = ref [] in
    (fun s -> cmdl := s :: !cmdl),
    (fun () -> List.rev !cmdl) in
  let arg_spec = Arg.align [
    ("-conf", Arg.String (fun x -> Config.set_conf_file (Some x)),
      (Printf.sprintf
        "FILE Sets the configuration file to use, if set to \"-\"\n\
         \t the configuration will be read from standard input."));
    ("-timeout", Arg.Int Config.set_timeout,
      (Printf.sprintf
        "INT Sets the maximal time waited by sklrun before deciding\n\
         \t that the computation network is down.\n\
         \t Default is %d."
        (Config.get_timeout ())));
    ("-err_len", Arg.Int Config.set_err_length,
      (Printf.sprintf
        "INT Sets the size of error messages given when the parser fails.\n\
         \t Default is %d."
        (Config.get_err_length ())));
    ("-warn_error", Arg.Bool Config.set_warn_error,
      (Printf.sprintf
        "BOOL Sets a flag which will stop sklrun if a warning occurs\n\
         \t during configuration processing.\n\
         \t Default is %B."
        (Config.get_warn_error ())));
    ("--", Arg.Rest append_cmd,
      (Printf.sprintf
        " Remainings arguments are treated as the Sklml command."));
  ] in
  let fail m =
    Printf.eprintf "failure: %s\nSee help using -help option\n" m;
    exit 1 in
  let head_usage =
    "usage: sklrun [OPTIONS] -- COMMAND\n" in
  try
    Arg.parse_argv argv arg_spec append_cmd head_usage;
    let cmd = get_cmd () in
    if cmd = [] then failwith "empty command" else cmd
  with
  | Failure m -> fail m
  | Arg.Bad _ -> fail "invalid argument"
  | Arg.Help m -> print_string m; exit 0
;;

let get_conf fname =
  let d =
    match fname with
    | None -> Right None
    | Some fname ->
      if fname = "-" then Right (Some stdin) else
      try Right (Some (open_in fname))
      with Sys_error m -> Left m in
  match d with
  | Right (Some ic) ->
    let rec loop () =
      let line =
        try Some (input_line ic)
        with End_of_file -> None in
      match line with
      | Some l -> l :: loop ()
      | None -> [] in
    Right (String.concat "\n" (loop ()))
  | Right None -> Right ""
  | Left _ as err -> err
;;

let main () =
  let cmd = parse_argv Sys.argv in
  match get_conf (Config.get_conf_file ()) with
  | Left err ->
    Printf.eprintf "Cannot retreive configuration: %s\n" err;
    exit 1
  | Right conf ->
    let warn m =
      Printf.eprintf "Warning, %s\n" m;
      if Config.get_warn_error () then begin
        Printf.eprintf "warn_error enabled, stopping here.\n";
        exit 1
      end in
    let mk_cmd addr port =
      let master = Printf.sprintf "%s:%d" addr port in
      let ec =
        Array.of_list
          (cmd @ [ "-master"; master; ]) in
      { Batch.program = ec.(0); Batch.args = ec; } in
    let local_skl_path = List.hd cmd in
    let kill_and_wait p =
      Unix.kill p Sys.sigterm; ignore (Unix.waitpid [] p) in
    try
      let b = mk_batch warn conf local_skl_path mk_cmd in
      match Batch.eval_batch b stdout stderr with
      | Batch.Ok pidl ->
        List.iter (fun pid -> ignore (Unix.waitpid [] pid)) pidl
      | Batch.Error (pidl, _) ->
        Printf.eprintf
          "There were errors, killing %d jobs.\n"
          (List.length pidl);
        List.iter kill_and_wait pidl
        (*List.iter (Printf.printf "\t%d\n") pidl*)
    with
    | Parse_error e ->
      Printf.eprintf "Error, parsing error in config near:\n%s\n\n" e
    | Expand_error e ->
      Printf.eprintf "Error, invalid format: %S\n\n" e
;;

let () = main ();;
